<?php
##### rrmodcompare_modlist.php, installed by ###Mod Manager Compare, by Robin Richmond #####
######### ALTERNATIVE VERSION THAT DOES NOT TRY TO READ admin_modhandler.php #######
#########       AND DOES NOT FIND MOD INSTALLATION STATUS #######
/* This version of the mod comparison server-side component reads .cfg files directly,
capturing the %modname, %version, %wikipage, %author, and %private tags. It also
reads the Mod Manager mod list (admin_modhandler.php) to get the installed status
from the HTML of the mod list.
Like all of the server-side components, it eliminates mods:
- whose modnames do not match the include or exclude strings,
- whose authors do not match the author search string (if there is an author), and
- whose private status does not match the "search private" flag.
It echoes data in JSON format.

The querystring parameters are:
- $searchauthor, which may be empty
- $authoryesno - 'y' if the author must match, 'n' if the author must NOT match, and
           '' if we're just highlighting the author. Relevant only if $searchauthor has a value
- $searchinclude - A mod name substring to match mods we want
- $searchexclude - A mod name substring to match mods we don't want
- $searchprivate - Truthy if we should include mods marked "private"
*/
include 'begin.php';

	$folder = $rootpath . $modspath;
	$modlist = array();
	$html = '';
	$handle = opendir($folder);
	while (false !== ($file = readdir($handle))) {
		if (is_dir("$folder/$file")) continue;
		if (substr($file,-4) != '.cfg') continue;
		#getCfgData returns a vector;
		# 0 = sortable modname
		# 1 = sortable version
		# 2 = sortable filename
		# 3 = associative array of data describing this mod.
		$modvector = getCfgData($file, ''); #no installation status
		if (is_array($modvector))
			$modlist[$modvector[0]][$modvector[1]][$modvector[2]] = $modvector[3];
	} #end of loop through the .cfg files

###### Sort $modlist and eliminate duplicates
# Sort it to make sure that we get the mods in case-insignificant modname order.
# We want only the highest-numbered version of each mod name.  That allows us
# to eliminate two subscripts from $modlist.
if (is_array($modlist)) {
	ksort($modlist);
	$newmodlist = array();
	foreach ($modlist as $modname => $level2) {
		#We only want the highest version #
		$v = '';
		foreach ($level2 as $versionnum=>$level3) {
			if ($versionnum>$v) {
				$v=$versionnum;
				$l=$level3;
			}
		}
		#Level 3 is a tiebreaker; take the last filename, too
		$f = '';
		foreach ($l as $filename=>$mydata)
			$lastfilename = $filename;
		$newmodlist[] = $mydata;
	} #end of top level of the 3-level loop through the modlist
	echo json_encode(array('numfiles'=>count($htmlmods), 'modlist'=>$newmodlist));
} else
	echo json_encode(array('numfiles'=>0, 'modlist'=>array()));
exit;

###########################################################################
#From this file, get the modname, author, version,  wikpage, and private attributes
#Also grab the installed status from the $installedLookup array that was read from a
#Mod Manager modlist.
function getCfgData($filename, $installedstatus) {
	global $searchprivate, $searchauthor, $authoryesno;
	global $installedLookup; # $installedLookup[{filename}]={installed status}
	$authorNames = array();
	$tagname = '';
	if ($searchauthor && $authoryesno=='y') $foundauthor = false;
	else $foundauthor = true;

	$modname = $version = $wikipage = $author = $authortext = $private = '';
	#Read the .cfg file looking for %modname, %version, %wikilink, and %author tags
	#Quit if we encounter a %target tag.
	$filespec = "{$GLOBALS['modspath']}/$filename";
	$cfghandle = fopen($filespec, 'r');
	$cfgline = rtrim(fgets($cfghandle),"\n\r");
	while ($cfgline!==false && substr($cfgline, 0, 7) != '%target') {
		if (substr($cfgline, 0,1) == '%') {
			$tagandparams = explode('%', $cfgline);
			$tagparams = explode(':', $tagandparams[1], 3);
			$tagname = $tagparams[0];
			switch ($tagname) {
				case 'target' : break 2; #exit the loop through the file
				case 'name' : $modname = $tagparams[1];
					if (!checkmodname($modname))
						return false; #Go on to next file
					break;
				case 'version' : $version = $tagparams[1];
						break;
				case 'wikipage' : $wikipage = $tagparams[1];
						break;
				case 'author' :
					if ($searchauthor && $authoryesno == 'n' && stripos($tagparams[1], $searchauthor)===0)
						return false;
					if (!$foundauthor && $searchauthor && $authoryesno == 'y' && stripos($tagparams[1],$searchauthor)===0)
						$foundauthor = true;
					$authorNames[] = $tagparams[1];
					#If there is an URL, create a hyperlink.
					if (count($tagparams)>2)
						$authorLinks[] = "<a href=\"{$tagparams[2]}\" target='_blank'>{$tagparams[1]}</a>";
					else #No URL, just save the author name in the list of hyperlinks
						$authorLinks[] = $tagparams[1];
					break;
				case 'private' : $private = 'private';
					if (!$searchprivate) return false; #Go on to next file
					break;
				default:
			} #end switch
		} #end if line starts with '%'
		$cfgline = rtrim(fgets($cfghandle),"\n\r");
	} #end loop through the cfg file
	fclose($cfghandle);

	#Handle the list of authors
	if (count($authorNames)) {
		#If we are searching for an author, that author must have already been found.
		if (!$foundauthor && $searchauthor && $authoryesno == 'y')
			return false;
		#Extract the author name from the first |-part of the first author
		$author = getFirstAuthor(strtok($authorNames[0], '|'));
		#The authortext is all author hyperlinks as a comma-space-delimited string.
		$authortext = implode(', ', $authorLinks);
	} #end there is at least one author tag

	#We have all of the .cfg file tag values we need. Return a vector with
	#all of those values (plus some values derived from these), indexed by
	# 1. Lowercase mod name, stripped of underscores, dashed, spaces, and a trailing "mod"
	# 2. Mod version number, in its sortable form
	# 3. Lowercase filename
	$lcmodname = strtolower(preg_replace('@[-_ ]@', '', $modname));
	$lcfilename = strtolower(preg_replace('@[-_ ]@', '', $filename));
	$versionPieces = getVersionWithoutVAndSortableVersion($version);
	#Return an array that will be added to the modlist
	return array($lcmodname, $versionPieces[1], $lcfilename,
		array('modname' => $modname,
			'modnamecompare' => $lcmodname,
			'version' => $versionPieces[0],
			'sortableVersion' => $versionPieces[1],
			'filename' => $filename,
			'wikipage' => $wikipage,
			'installed' => $installedstatus,
			'author' => $author,
			'authortext' => $authortext,
			'private' => $private));
} #end function getCfgData

###############################################################
#Try to find a name within the first 3 words of authortext
function getFirstAuthor($authortext) {
	#Truncate any text that follows a comma or semicolon
	$authortext = explode(';', explode(',', $authortext)[0])[0];
	$words = explode(' ',$authortext, 4);
	if (count($words)==4) array_pop($words);
	#Assume that the 2nd & 3rd words are not part of the name if they don't start
	#with a capital letter.
	if (count($words)>1) {
		if (!ctype_upper(substr($words[1], 0, 1)))
			#Get rid of words 2 and 3; we're left with 1 word.
			$words = array($words[0]);
		elseif (count($words)==3 && !ctype_upper(substr($words[2], 0, 1)))
			array_pop($words); #We're left with just 2 words.
		#Keep the third word only if the second word is just an initial.
		if (count($words)==3 && !preg_match('@^\w+\.?$@' ,$words[1]))
			array_pop($words);
	}
	return implode(' ',$words);
} #end function getFirstAuthor

###############################################$#
# If there is an include string, the modname must include it.
# If there is an exclude string, the modname must NOT include it.
# If a search string starts and ends with '/', it is treated as a regular expression.
# Return false if the modname does NOT qualify.
# Ignore spaces in the modname and in the search strings.
function checkmodname($modname) {
	global $searchinclude, $searchexclude;
	$comparemodname = str_replace(' ', '', strtolower($modname));
	#Check for string that must be in the mod name
	if ($searchinclude) {
		$searchinc = str_replace(' ', '', strtolower($searchinclude));
		if (substr($searchinclude,0,1)=='/' && substr($searchinclude,-1,1)=='/') {
			if (!preg_match($searchinc.'im', $comparemodname))
				return false; #Skip this mod
		} else {
			if (stripos($comparemodname, $searchinc) === false)
				return false; #Skip this mod
		}
	}
	#Check for string that must not be in the mod name
	if ($searchexclude) {
		$searchexc = str_replace(' ', '', strtolower($searchexclude));
		if (substr($searchexclude,0,1)=='/' && substr($searchexclude,-1,1)=='/') {
			if (preg_match($searchexc.'im', $comparemodname))
				return false; #Skip this mod
		} else {
			if (stripos($comparemodname, $searchexc) !== false)
				return false; #Skip this mod
		}
	}
	return true;
} #end of checkmodame

########################################################
#Check the author name to see if we can skip this mod
function checkauthor($author) {
	global $searchauthor, $authoryesno;
	if ($author) {
		$authormatches = (strpos($author, $searchauthor)!==false);
		if (($authoryesno=='y' && !$authormatches) ||
			($authoryesno=='n' && $authormatches))
		return false;
	}
	return true;
} #end function checkauthor

###############################################
#Extract the version # & its sortable variant from a Mod Manager results table
function getVersionWithoutVAndSortableVersion($version) {
	#Trim the letter v from the version #
	$version = trim($version, "Vv");
	if (preg_match('@(\d+\.\d\.\d+\.)(\d+)(.*)@si', $version, $matches)) {
		#Use the buildSortableVersion function to pad the version # so it can
		#be sorted
		$sortableVersion = buildSortableVersion($matches);
		return array($version, $sortableVersion);
	} else #no valid version number
		return array('??','??');
} #end function getVersionWithoutVAndSortableVersion

#########################################################
# Convert a parsed version # so that it can be sorted.
# The input parameter is a match array from preg_replace
# The first matched substring is the three dot parts of the TNG version #
# The second matched substring is the numeric part of the 4th dot-piece.
# The third matched substring contains any alphanumeric suffix in the 4th dot-piece
function buildSortableVersion($matches) {
	#Get the first 3 parts of the version #
	$tngVersion = $matches[1];
	#Add a zero if the first digit of the major TNG version is not '1'
	if (substr($tngVersion,0,1)<>'1') $tngVersion = "0$tngVersion";
	#Get the numeric part of the last of the 4 dot-pieces) and pad it with zeroes
	$lastNumber = sprintf("%03d",$matches[2]);
	#Put the 3 parts back together
	return $tngVersion . $lastNumber . $matches[3];
} #end function buildSortableVersion
?>