<?php
##### rrmodcompare_report.php, installed by ###Mod Manager Compare, by Robin Richmond
/* Prints a mod comparison report, comparing local mods to
a) The mods in a second TNG site and/or
b) The mods' Wiki articles.
It
- Uses file_get_contents to capture the Mod Manager List web page (admin_modhandler.php)
  from each environment.
- Parses the HTML results table from each environment to extract information about
  each mod.
- Optionally, uses file_get_contents to capture each mod's Wiki article, and extract
  more information about the mod.
A kickoff form (in rrmodcompare_report) provide options that drive this report

The options are (approximately):
- $site2url, the site2 URL, which may be empty
- $wiki, compare to the wiki; (checkbox; 1 or missing)
- $author, which may be empty
- $sinclude - A mod name substring to match mods we want
- $sexclude - A mod name substring to match mods we don't want
- $ayesno - 'y' if the author must match, 'n' if the author must NOT match, and
           '' if we're just highlighting the author. Relevant only if $author has a value
- $private - True if we are accepting private mods.
- $site1vwikifield[] - array of checkboxes whose values qualify the mod to be displayed
- $site2vwikifield[] - likewise
- $site2vsite1field[] - likewise
- $wikifield[] - likewise
- $sb: - radio button: values represent mod name, filename, author, filename, and status options
- $life & $unit - Cache life (Integer of weeks, days, hours,
         or minutes before the cache (for a given Wiki article) expires
- $cachemethod - radio button values 1-3
- $whichoptions - 'advanced', 'basic', or '' (for the one default search condition)

Overall, the search criteria can be expressed this way:
  The mod is accepted if
         there is an author search key, and it is supposed to match the author name(s), and it does.
    AND, if there is an author search key and it is NOT supposed to match the author, it doesn't.
    AND, if there is an "include string", it "matches" the modname,
    AND, if there is an "exclude string", it does NOT match the modname.
	AND, if the "Private" parameter is false, the mod is not private.
    AND, if there are any status search values, at least one of them matches the corresponding status value.

Note that a "match" on the author name(s) means that the author search key is a substring in the list of authors.  The author list come first from the .cfg file, and if there are no authors defined in the .cfg file, the author list comes from the Wiki article.

A match against the modname can be a regular match or a substring match.
If the "include string" or "exclude string" starts and ends with "/", then it is a regular expression pattern. With regular expressions, "match" means that a caseless preg_match with the "include string" must succeed, and a caseless preg match with the "exclude string" fails. With ordinary strings, "match" means that the string must be (or not be) a substring within the mod name.

Regarding status values, there are up to 6 status columns in the results:
  1. The site 1 installation status,
  2. The site 2 installation status
  3. The "site 1 vs wiki" status (which is mostly a version # comparison
  4. The "site 2 vs wiki" status
  5. The "site 2 vs site 1" status (which is also a version # comparison)
  6. The wiki status (which is based on whether there is a wiki article, whether there is
     a version number, and whether there is more than one version number)
A status value search can be conducted on the latter 4 status columns only (i.e. not the installation statuses). The kickoff form presents checkboxes for each possible value of those 4 columns (if the column is relevant), and the user may may check any number of status values.  If there are status values, then at least one of those status values must appear in the corresponding column (or field) for each mod.

The search for acceptable mods is not conducted with a single logical expression.
It works kind of like this:
1. Read each mod .cfg file on site 1, and, for the highest version # of each mod name:
   a. The modname must match the include string and must NOT match the exclude string.
   b. If there are author names, and there is an author search string, and the
      "ayesno" parameter is "y" or "n", then the author names must either contain
	  or not contain (depending on "ayesno") the author search string.
   c. If there is an include string, it must be present, or must match.
   d. If there is an exclude string, it must NOT be present, or must NOT match.
   e. If the "Private" flag is off, then the mod must not be private.

2. If a second TNG site is specified, read each mod .cfg file on site 2, and,
   for the highest version # of each mod name, do the same 5 tests as on site 1.

3. Next, walk through the combined list of mods. For each mod:
   a. Read the Wiki article. (If there is no Wiki article; that's ok for now.)
   b. If no authors were defined in the mod's cfg file,
      and there is a list of authors in the Wiki article,
	  and there is an author search string'
	  and the "ayesno" parameter is "y" or "n"
	  Then the author search string must appear (or not appear) in the list of authors.
	  If it doesn't, reject the mod.
   c. Compute the Site1 vs Wiki, Site2 vs Wiki, Site2 vs Site1, and Wiki statuses.
   d. Generate a HTML string that represent the result table row for this mod.
   e. Return that HTML string, plus the 4 status values
   f. If there are status search values, and none of them appear, reject the mod.
*/
#Set the flag that lets this program read the Mod Manager Compare language strings.
$mod_manager_compare_cust_text = 'report';
$tngModNames['Mod Manager Compare']['rrmodcompare_report.php'] = '7';
@ini_set( "allow_url_fopen", "1" ); #we evidently need this PHP setting that globallib.php resets
####################################################################

#These $rrconfig settings suppress the Admin Noframes Menu
$rrconfig['NoFramesShowLeft'] = 0;
$rrconfig['NoFramesShowRight'] = 0;

if (empty($site2url)) $site2url = '';

#Pattern to extract the mod version number from a string such as a mod filename
#or the Latest Version section of the Wiki article sidebar
define('VERSIONPATTERN','@v*(\d+\.\d\.\d+\.)(\d+)([-a-z0-9]*)@si');
# Define values used by function set_horizontal_tabs
define( 'YES',1 );
define( 'NO', 0 );

include("begin.php");
include 'adminlib.php';
$textpart = "mods";
include("$mylanguage/admintext.php");

########################################################
#Read custom language strings from a mods subfolder rather than from regular cust_text.php files.
$modfolder = $admtext['mmmodfolder'];
if (file_exists("$modspath/$modfolder/{$mylanguage}_cust_text.php"))
	include("$modspath/$modfolder/{$mylanguage}_cust_text.php");
else include("$modspath/$modfolder/languages/English_cust_text.php");

# Don't require admin login
$admin_login = 0;
include 'checklogin.php';
include 'version.php';
include 'classes/version.php';
@include ("$extspath/wikicache.php");
$helplang = findhelp("rrmodcompare_help.php");
$modspath = rtrim($modspath,'/');

########## Mod Comparison Report Options Setup ##########
#All of the report's options are saved in $options[] within the program;
#they are not saved to the mod manager config file.

# Define the options that determines which columns are displayed
$options['rt'] = isset($rt) ? $rt : 1;
$options['reportclass'] = ($rt > 1);
$options['fromform'] = (isset($b1) && $b1); #Based on the submit button

$options['needupdate'] = isset($needupdate) ? $needupdate : 0;
if ($options['needupdate']) $wiki=1;
$options['wiki'] = isset($wiki) ? $wiki : 0;

#Cache method & parameters
if (!isset($cachemethod) || !$cachemethod) $cachemethod = 3;
$options['cachemethod'] = $cachemethod;
$options['life'] = isset($life) ? $life : 5;
$options['unit'] = isset($unit) ? $unit : 'days';

$options['sinclude'] = isset($sinclude) ? $sinclude : '';
$options['sexclude'] = isset($sexclude) ? $sexclude : '';
$options['private'] = isset($private) ? $private : 0;
$options['author'] = isset($author) ? $author : '';
$options['ayesno'] = isset($ayesno) ? $ayesno : '';

#### Display Options
$options['sb'] = isset($sb) ? $sb : 'modname'; #Sort
$options['aformat'] = isset($aformat) ? $aformat : 'initial'; #author format
$options['wrap'] = 1; #Can wrap filename; not set by kickoff form any more
$options['statushilite'] = isset($statushilite) ? $statushilite : 0;
$options['hideversion'] = isset($hideversion) ? $hideversion : 0;
#Flag mismatches - whe filenames & Wiki titles don't match the mod name or v#
$options['mismatch'] = isset($mismatch) ? $mismatch : 0;
#Show superscripts for users found in the Wiki, and for Wiki articles under construction
$options['showsuper'] = isset($showsuper) ? $showsuper : 0;
#Show a column for cache age
$options['showcache'] = isset($showcache) ? $showcache : 0;
#Move author to the far right
$options['aright'] = isset($aright) ? $aright : 0;

if ($options['fromform']) {
	################ Submitted through the kickoff form ################
	#Save cookies for each input parameter
	$exptime = 0; #Cookie expiration date - all cookies are session cookies

	if (!isset($power)) $power = '';
	#Site 2
	$site2url=trim($site2url);
	#Make sure that the Site 2 URL ends with / and begins with http:// or https://
	if ($site2url) {
		if (substr($site2url,-1) != '/') $site2url .= '/';
		if (substr($site2url,0,7)!=='http://' && substr($site2url,0,8)!='https://')
			$site2url = 'http://' . $site2url;
	}
	$options['site2url'] = $site2url;
	$options['orphans'] = isset($orphans) ? $orphans : 0;
	#By default, internal reports do not show the mod summary
	$options['summary'] = isset($summary) ? $summary : 0;

	setcookie("tng_modcompare[power]", $power, $exptime);
	setcookie("tng_modcompare[rt]", $options['rt'], $exptime);
	setcookie("tng_modcompare[wiki]", $options['wiki'], $exptime);
	setcookie("tng_modcompare[needupdate]", $options['needupdate'], $exptime);
	setcookie("tng_modcompare[cachemethod]", $options['cachemethod'], $exptime);
	setcookie("tng_modcompare[cachemethodused]", $options['cachemethod'], $exptime);
	setcookie("tng_modcompare[life]", $options['life'], $exptime);
	setcookie("tng_modcompare[unit]", $options['unit'], $exptime);
	setcookie("tng_modcompare[sinclude]", $options['sinclude'], $exptime);
	setcookie("tng_modcompare[sexclude]", $options['sexclude'], $exptime);
	setcookie("tng_modcompare[private]", $options['private'], $exptime);

	setcookie("tng_modcompare[site2url]", $options['site2url'], $exptime);
	setcookie("tng_modcompare[orphans]", $options['orphans'], $exptime);
	setcookie("tng_modcompare[author]", $options['author'], $exptime);
	setcookie("tng_modcompare[ayesno]", $options['ayesno'], $exptime);

	setcookie("tng_modcompare[sb]", $options['sb'], $exptime);
	setcookie("tng_modcompare[aformat]", $options['aformat'], $exptime);
	setcookie("tng_modcompare[statushilite]", $options['statushilite'], $exptime);
	setcookie("tng_modcompare[summary]", $options['summary'], $exptime);
	setcookie("tng_modcompare[wrap]", $options['wrap'], $exptime);
	setcookie("tng_modcompare[hideversion]", $options['hideversion'], $exptime);
	setcookie("tng_modcompare[mismatch]", $options['mismatch'], $exptime);
	setcookie("tng_modcompare[showsuper]", $options['showsuper'], $exptime);
	setcookie("tng_modcompare[showcache]", $options['showcache'], $exptime);
	setcookie("tng_modcompare[aright]", $options['aright'], $exptime);

} else {
	########### We were not submitted from the kickoff form #########
	#Don't set cookies, and hard-code several options
	$options['rt'] = 1;
	$options['reportclass'] = 0;
	$options['wiki'] = 1;
	$options['site2url'] = '';
	#By default, the external report shows the mod summary
	$options['summary'] = isset($summary) ? $summary : 1;
	#By default, put the author column on the right.
	$options['aright'] = isset($aright) ? $aright : 1;
	#Options that are used only in external reports
	if (isset($installed)) $installed = $installed;
	elseif (isset($inst)) $installed = $inst;
	else $installed = 0;
	$options['installed'] = $installed;
	$options['v'] = isset($v) ? $v : 0;
} #end if externally lauched or submitted from kickoff

#Derived parameters
$options['authorisrelevant'] = ($options['author'] && $options['ayesno'] <> '');
$options['regexinclude'] = (substr($options['sinclude'],0,1) == '/' && substr($options['sinclude'],-1,1) == '/');
$options['regexexclude'] = (substr($options['sexclude'],0,1) == '/' && substr($options['sexclude'],-1,1) == '/');

#Using the cache limit and time unit, calculate the cutoff date.
#Any cache entries saved before this cutoff date have expired.
if ($options['life']>0) {
	if ($options['unit']=='minutes') $seconds = 60;
	elseif ($options['unit']=='hours') $seconds = 3600;
	elseif ($options['unit']=='weeks') $seconds = 7*86400;
	else $seconds = 86400; #Default to days
	$options['cachecutoff'] = time() - $options['life']*$seconds;
} else $options['cachecutoff'] = 0; #Everything has expired

#To facilitate the process of checking each mod against the list of status value search keys,
#store the form field querystring parameters in arrays that are indexed by the selected
#status search keys, and create a boolean variable that tells us if there are any search keys.
$options['statusconditions'] = false; #updated as a side effect of createStatusLookupArray()
createStatusLookupArray('site1vwiki');
createStatusLookupArray('site2vwiki');
createStatusLookupArray('site2vsite1');
#Setup a status search value implied by $needupdate
#(That is,"search for mods that have a newer verion on the Wiki")
if ($options['needupdate']) {
	$options['statusvalues']['site1vwiki']['<wiki'] = 1;
	$options['statusconditions'] = true;
}

################ End of cookie and option setup ################

######### HTML STREAM STARTS HERE #############
$flags['tabs'] = $tngconfig['tabs'];
# Don't use Mod Manager styles $flags['modmgr'] = true;
tng_adminheader( $admtext['modmgr'], $flags );
#Try to make sure that the user's browser keeps up with status information displayed by this program.
@ob_implicit_flush(true);
@ob_end_flush();
?>
<script>
$(function() {
	//////////////////////////////////////////////////
	//Insert this program's standalone style sheet in the DOM before genstyle.css so
	//modifications to this stylesheet can be made in genstyle.css.
	$("link[href^='css/genstyle.css']").before("<link rel='stylesheet' type='text/css' href='css/rrmodcompare.css'/>");

	//Apply the "don't print on printer" class to the tab menu and innermenu
	$('#adm-innermenu').parent().addClass('rrmcnoprint');

	//Apply the 'fieldname' class to all of the th cells in the results table
	$('#results th').addClass('fieldname');

	//Erase the list of mods that that were displayed as we loaded and processed them
	$('#initialcounts').hide();

	//Wiki summary toggle
	$('.summarytoggle').click(function (){
		$(".summarytoggle,.summarytoggle2").toggle();
	});

<?php
#Apply striping to the results table rows if the striping flag is set
if ($options['use_striping']) {
	$n = 2*$options['stripe_after'];
	for ($ii=1;$ii<=$options['stripe_after'];$ii++) {
		$iii=$ii+3;
	echo "$('#results tbody tr:nth-child({$n}n+{$ii})').addClass('databack');\n";
	echo "$('#results tbody tr:nth-child({$n}n+{$iii})').addClass('databackalt');\n";
	}
}
?>
}); //End of (document).ready function
</script>
<script type='text/javascript' src='js/admin.js'></script>
</head>

<?php
if ($options['fromform']) {
	#Normal TNG heading when called from form
	$reporttitle = "{$admtext['modmgr']} >> {$admtext['mmcomparereport']}";
	echo tng_adminlayout();
} else { #Contrived heading when not called from form
	if ($ayesno == 'n')
		$subtitle = $options['comparesubtitleother'];
	elseif ($ayesno == 'y')
		$subtitle = $options['comparesubtitlemine'];
	else $subtitle = $options['comparesubtitleany'];
	$reporttitle = "<h2>$sitename - Mods</h2>";
	$reporttitle .= "<h3>$tngdomain</h3>";
	if ($subtitle) $reporttitle .= "<h4>$subtitle</h4>";
	$reporttitle = "<table class='rrmymodstitle normal'><tr><td>" . nl2br($reporttitle) . "</td></tr></table>";
}
if ($options['fromform']) {
	#The form was submitted; display tabs and innermenu
	$modtabs = set_horizontal_tabs( $options['show_analyzer'], $options['show_developer'], $options['show_updates']);
	#Inner menu link to Mod Manager Comparison Report help
	$innermenu = "<a href='$helplang/rrmodcompare_help.php' onclick=\"return openHelp('$helplang/rrmodcompare_help.php');\" class='lightlink'>{$admtext['mmcomparehelp']}</a>";
	# link to the Mod Comparison wiki article.
	$innermenu .= " &nbsp | &nbsp; <a href='https://tng.lythgoes.net/wiki/index.php/Mod_Manager_Compare' onclick=\"return openHelp('https://tng.lythgoes.net/wiki/index.php/Mod_Manager_Compare');\" class='lightlink'>{$admtext['mmcomparewiki']}</a>";
	# link to the Report Legend section of the Wiki article.
	$innermenu .= " &nbsp | &nbsp; <a href='https://tng.lythgoes.net/wiki/index.php/Mod_Manager_Compare#Column+Headings' onclick=\"return openHelp('https://tng.lythgoes.net/wiki/index.php/Mod_Manager_Compare#Column+Headings');\" class='lightlink'>{$admtext['mmlegend']}</a>";
	$menu = "<div class=\"mmmenuwrap\">";
	$menu .= doMenu($modtabs,"compare",$innermenu);
	$menu .= "</div>\n";
} else $menu = ''; #External call, no forms or tabs

if(!isset($message)) $message = "";
$headline = displayHeadline($reporttitle, "img/modmgr_icon.gif", $menu, $message) . "\n";
echo "<div id=\"mmhead\" class=\"$headclass adminback\">
	$headline
</div><!--head-section-->\n";
#echo str_replace("\n<br />", "", $html); #get rid of some excess whitespace
?>
<table width='100%' border='0' cellpadding='3' class='lightback'>
	<tr class='databack'>
	<td class='tngshadow normal'>
<?php
foreach (explode(',', 'fromform,authorisrelevant,statusoptions,sinclude,sexclude,private') as $option)
	if (!isset($options[$option])) $options[$option] = '';

########################## Main Program ###############################
if ($options['fromform']) {
	####Display the report options
	#Author highlighted but not required:
	if (!$options['authorisrelevant'] && $options['author'])
		echo "<div><b>{$admtext['mmauthorhigh']}</b>: {$options['author']}</div>\n";

	#Are there any restrictive options?
	if ($options['authorisrelevant'] || $options['statusoptions'] || $options['sinclude'] ||  $options['sexclude'] || !$options['private']) {
		$connector = '';
		echo "<div class='rrlefthanging'><b>{$admtext['mmsearchcriteria']}</b>:";
		$criterionSeen = false;
		#Author
		if ($options['authorisrelevant']) {
			$criterionSeen = true;
			if ($options['ayesno']=='y')
				echo "<br>{$admtext['mmmodswritten']}";
			elseif ($options['ayesno']=='n')
				echo "<br>{$admtext['mmnotwritten']}";
			echo " {$options['author']}";
			$connector = "{$admtext['mmand']} ";
		}
		#Include or exclude based on string
		if ($options['sinclude']) {
			$criterionSeen = true;
			$langstring = $options['regexinclude'] ? 'doesmatch' : 'doesinclude';
			echo "<div class='rrlefthanging'>{$connector} {$admtext[$langstring]} '{$options['sinclude']}'</div>\n";
			$connector = "{$admtext['mmand']} ";
		}
		if ($options['sexclude']) {
			$criterionSeen = true;
			$langstring = $options['regexexclude'] ? 'notmatch' : 'notinclude';
			echo "<div class='rrlefthanging'>{$connector} {$admtext[$langstring]} '{$options['sexclude']}'</div>\n";
			$connector = "{$admtext['mmand']} ";
		}
		#Exclude private mods
		if (!$options['private']) {
			$criterionSeen = true;
			echo "<div class='rrlefthanging'>{$connector} {$admtext['notprivate']}</div>\n";
			$connector = "{$admtext['mmand']} ";
		}
		#Status values:
		if ($options['statusconditions']) {
			$criterionSeen = true;
			echo "<div class='rrlefthanging'>{$connector} {$admtext['modswithstatus']}:";
			echo displaySelectedOptions('site1vwiki');
			echo displaySelectedOptions('site2vwiki');
			echo displaySelectedOptions('site2vsite1');
			echo "</div>\n"; #end the status values div
		}
		if (!$criterionSeen) echo $text['none'];
		echo "</div>\n"; #Close the div that wraps and indents the restrictive options
	}
	#Sort options
	if ($options['fromform'] && $options['sb']!='modname') {
		$sortbydisplayname = $admtext['mmsort'.$options['sb']];
		echo "<div><b>{$admtext['sortby']}</b>: $sortbydisplayname</div>\n";
	}
	#Cache options
	if ($wiki) {
		echo "<div><b>{$admtext['mmcachemethod']}:</b> ";
		if ($options['cachemethod']==1) echo "{$admtext['mmignorecache']}</div>\n";
		elseif ($options['cachemethod']==2)
			echo str_replace('XXX', "{$options['life']} {$options['unit']}", $admtext['mmusecacheif']) . "</div>\n";
		else echo "{$admtext['mmalwayscache']}</div>\n";
	}
	########## End of information displayed above the results table
} #End if invoked from form

	$rowcounts = array();  #Initialize counts of mod status
	foreach (explode(',', 'all,site1analyzed,wikicache,wikiserver,wikiread,wiki,author') as $ss)
		$rowcounts[$ss] = 0;
	#We'll count status values for each status column
	$rowcounts['site1vwiki'] = array();
	$rowcounts['site2vwiki'] = array();
	$rowcounts['site2vsite1'] = array();
	$rowcounts['wiki'] = array();

	#######################
	### Parse the local and possibly remote Mod Manager cfg files into the numbered arrays
	#$site1mods and $site2mods.
	# Each numbered element of $site{i}Mods is an associative array whose top level is
	# a numbered array (a vector), and the data is stored with indexes such as
	# 'modnamecompare', 'modname', 'version', 'sortedversion', 'filename', 'wikilink',
	# 'installed', 'author', 'private'
	$options['site1url'] = $tngdomain;
	if (substr($options['site1url'],-1) != '/') $options['site1url'] .= '/';

	$results = readCfgFileDataFromSite($options['site1url']);
	$rowcounts['site1files'] = $results['numfiles'];
	$site1mods = $results['modlist'];
	if (is_array($site1mods) && count($site1mods)) {
		$site1count = count($site1mods);
		if ($options['fromform']) {
			echo "<br><b>{$admtext['readingcfg']} {$options['site1url']}&hellip; </b>";
			echo +$rowcounts['site1files'],' ',$admtext['filesseen'], '; ', +$site1count, " {$admtext['mmmodsmatch']}\n";
		}
		if ($options['site2url']) {
			echo "<br><b>{$admtext['readingcfg']} {$options['site2url']}&hellip; </b>";
			#Now the Site2 mod data
			$results = readCfgFileDataFromSite($options['site2url']);
			$rowcounts['site2files'] = $results['numfiles'];
			$site2mods = $results['modlist'];
			$site2count = count($site2mods);
			echo +$rowcounts['site2files'],' ',$admtext['filesseen'], '; ', +$site2count, " {$admtext['mmmodsmatch']}\n";
		} else { #No site2 address
			$site2mods = array();
			$site2count = 0;
		}
		#See if we found mods that we need to process
		if ($site1count || $site2count) {
			#Get the mod count by creating an associative array whose indexes are mod names
			if ($site1count && $site2count) {
				foreach ($site1mods as $modarray)
					$modlist[$modarray['modname']] =1;
				foreach ($site2mods as $modarray)
					if (!isset($modlist[$modarray['modname']])) $modlist[$modarray['modname']] = 1;
					else $modlist[$modarray['modname']]++;
				$modcount = count($modlist);
				#Say "Combined, there are XXX mods
				echo '<br>' . str_replace('XXX', $modcount, $admtext['combinedmods']);
			}
			elseif ($site1count)
				$modcount = $site1count;
			else $modcount = $site2count;

			#Start the display of wiki articles read from the wiki
			if ($options['fromform']&& $options['wiki'])
				echo "<br><b>{$admtext['mmreadingwiki']}&hellip;</b> <span id='initialcounts'>";

			#Combine the two sites into an array of table rows
			$modResultsList = getSortedResults ($site1mods, $site2mods);

			#Save the cache file if we read any wiki files.
			if ($rowcounts['wikiread'])
				saveCacheFile($wikicache);

			if ($options['fromform'] && $options['wiki']) {
				#Finish the description of the Wiki scan
				echo "</span>"; #End of the initialcounts span, which will be erased
				echo +$rowcounts['wikiread'], ' ', $admtext['mmwikiread'], '; ',
					 +$rowcounts['wikicache'], ' '. $admtext['mmpulledfrom'];
				$difference = $modcount - $rowcounts['wikiread'] - $rowcounts['wikicache'];
				if ($difference)
					echo "<span class='nw'>; $difference {$admtext['mmnowiki']}.</span>";
				echo "</div>\n";
			}
			#Display the count of mods that matched the author we were searching for.
			#Note that the author is not necessarily a filter. We may just be highlighting an author.
			if ($options['author'] && $rowcounts['author'])
				echo "<br><b>{$admtext['modsby']}</b> {$options['author']}: {$rowcounts['author']}\n";

			if (count($modResultsList)) {
				printResultsTable($modResultsList);
				if ($site2count || $options['wiki'])
					printSummaryCounts($site2count);
			} else
				echo "<p><b>{$admtext['nomodsmatch']}</b></p>\n";
		} #End if mods on site 1 or site 2
		else echo "<p><b>{$admtext['nomodsmatch']}</b></p>\n";

		if ($options['fromform']) 
			####### Link to run another report
			echo "<p><a href='rrmodcompare_form.php'>{$admtext['anothercompare']}</a></p>\n";
	} #End we found local files to process
	else echo "<p><b>{$admtext['nomodsmatch']}</b></p>\n";

?>
</td>
</tr>
</table>
<?php echo "<div align=\"right\"><span class=\"normal\">$tng_title, v.$tng_version</span></div>"; ?>
</body>
</html>
<?php
##################################################################
##################################################################
########## End of main routine; PHP functions follow #############
##################################################################

##################################################################
# Merge the arrays of local and remote mods into an array of result table rows.
# Return an array of HTML strings that represent rows in the results table, indexed by the sort key.
# Process:
# Step through the two arrays of mod descriptions, and call createResultTableRows to create a HTML string.
# The function createResultTableRows reads Wiki data only if the wiki flag is set.
# If the Wiki article for a mod failed the Author test, createResultTableRows returns an empty string.
# If we do get an HTML string, use checkSelectedOptions to see whether the status values in the row match
# the status key values.
# Only then will we add the this mod's data to the array of mod results.
function getSortedResults($site1mods, $site2mods) {
	global $options;
	$modResultsArray = array();
	$site1count = count($site1mods);
	$site2count = count($site2mods);
	$site1 = $site2 = 0; #Indexes within $site1mods and $site2mods
	$site1ModData = $site1count ? $site1mods[0] : '';
	$site2ModData = $site2count ? $site2mods[0] : '';

	while ($site1<$site1count && $site2<$site2count) {
		$combinedResult = array();
		$x = strcmp($site1ModData['modnamecompare'], $site2ModData['modnamecompare']);
		if (!$x) { #The two modnames are equal, so we have to merge data from the 2 sites.
			$combinedResult = createResultRow($site1ModData, $site2ModData);
			$site1ModData = $site1mods[++$site1];
			$site2ModData = $site2mods[++$site2];
		}
		elseif ($x < 0) { #site1 modname < site2 modname, so use site 1 only.
			$combinedResult = createResultRow($site1ModData, '');
			$site1ModData = $site1mods[++$site1];
		}
		else { #site1 modname > site2 modname, so use site 2 only.
			#A flag tells us whether to include site 2 orphans.
			if ($options['orphans'])
				$combinedResult = createResultRow('', $site2ModData);
			$site2ModData = $site2mods[++$site2];
		}
		##Look at the data in the merge result to see if we want to keep this mod.
		if (isset($combinedResult['tablerow']) && $combinedResult['tablerow'] && checkSelectedOptions($combinedResult))
			$modResultsArray[$combinedResult[$options['sb']]][$combinedResult['modname']] = $combinedResult;
	}
	# Gather site1 mods that come after the last site2 mod
	while ($site1 < $site1count ) {
		$combinedResult = createResultRow($site1mods[$site1++], '');
		if ($combinedResult['tablerow'] && checkSelectedOptions($combinedResult))
			$modResultsArray[$combinedResult[$options['sb']]][$combinedResult['modname']] = $combinedResult;
	}

	# Gather site2 mods that come after the last site1 mod, but only if we are processing orphans
	if ($options['orphans']) {
		while ($site2<$site2count) {
			$combinedResult = createResultRow('', $site2mods[$site2++]);
			if ($combinedResult['tablerow'] && checkSelectedOptions($combinedResult))
				$modResultsArray[$combinedResult[$options['sb']]][$combinedResult['modname']] = $combinedResult;
		}
	}
	ksort($modResultsArray);
	return $modResultsArray;
} #end function getSortedResults

#########################################################################
# Function to output the results table based on an associative array of table rows.
# The array of table rows is sorted by the table sort key.
function printResultsTable($modResults) {
	global $options, $admtext, $rowcounts;

	#Sort flags for when there is room to space it over, and when there isn't.
	$sortflag1 = "&nbsp;&nbsp;&nbsp;<img src='img/tng_sort_desc.gif' width='10' height='8' border='0' alt=''>";
	$sortflag2 = "<img src='img/tng_sort_desc.gif' width='10' height='8' border='0'";

	##### Display the column headings #####
	$tablehtml = "<table id='results' class='normal' cellspacing='0'>
	<thead>
	<tr style='vertical-align-bottom:' class='fieldnameback'>
		<th>#</th>\n";

	### Modname and Filename
	$sort1 = $options['sb']=='modname' ? $sortflag1 : '';
	$sort2 = $options['sb']=='filename' ? $sortflag1 : '';
	if ($options['reportclass']) {
		#normal report - both modname and filename
		$tablehtml .= "<th>{$admtext['modname']}$sort1</th>
			<th>{$admtext['filename']}$sort2</th>\n";
	} else { #simple report - use flag to decide between modname or filename
		if ($options['sb'] == 'filename')
			$tablehtml .= "<th>{$admtext['filename']}$sort1</th>\n";
		else $tablehtml .= "<th>{$admtext['modname']}$sort2</th>\n";
	}

	###### Author goes here unless the 'aright' flag is set
	$sort = $options['sb']=='author' ? $sortflag1 : '';
	if ($options['aformat']!='none' && !$options['aright'])
		$tablehtml .= "<th class='authorcol' style='min-width:55px;'>{$admtext['mmcauthor']}$sort</th>\n";

	###Border between identification and site 1, but only if there is a site 2
	if ($options['site2url'])
		$colstyle = 'rrseparator';
	else $colstyle = '';

	###Site 1 columns
	#Omit installation status when coming from outside unless the installed flag is set.
	if ($options['fromform'] || $options['installed']) {
		$tablehtml .= "<th class='$colstyle'>{$admtext['mmchinstall']}</th>\n";
		$colstyle='';
	}

	#No version when coming from outside unless the 'v' flag is set
	if ($options['fromform'] || $options['v']) {
		#Version # heading may say "Local Version" or just "Version"
		if ($options['site2url']) $prefix = 'L'; #local version
		else $prefix = '';
		$tablehtml .= "<th class='nw $colstyle'>" . str_replace(' ', '&thinsp;', $admtext["mmch{$prefix}version"]) ."</th>\n";
		$colstyle='';
	}

	#Site1 v Wiki - dependent on the Wiki & the 'v' flag 
	$sort = $options['sb']=='site1vwiki' ? $sortflag2 : '';
	if ($options['wiki'] && ($options['fromform'] || $options['v'])) {
		if ($options['site2url']) $prefix = 'L'; #local
		else $prefix = '';
		$tablehtml .= "<th class='$colstyle'>" . str_replace(' ', '&thinsp;', $admtext["mmch{$prefix}wiki"]) . "$sort</th>\n";
		$colstyle='';
	}
	###All Site 2 columns installed, v#, Site2VSite1, Sit32vWiki, - starting with a separator
	if ($options['site2url']) {
		$sort1 = $options['sb']=='site2vsite1' ? $sortflag2 : '';
		$sort2 = $options['sb']=='site2vwiki' ? $sortflag2 : '';
		$tablehtml .= "<th class='rrsite2data nw rrseparator'>" . str_replace(' ', '&thinsp;', $admtext['mmch2install']) . "</th>
		<th class='rrsite2data nw'>" . str_replace(' ', '&thinsp;', $admtext['mmch2version']) . "</th>
		<th class='rrsite2data nw'>" . str_replace(' ', '&thinsp;', $admtext['mmch2site1']) . "$sort1</th>\n";
		if ($options['wiki'])
			$tablehtml .= "<th class='rrsite2data rrwikidata nw'>" . str_replace(' ', '&thinsp;', $admtext['mmch2wiki']) . "$sort2</th>\n";
	}

	###Wiki columns - starting with a separator
	if ($options['wiki']) {
		$sort = $options['sb']=='date' ? $sortflag2 : '';
		if ($options['reportclass']) {
			#Full reporttype: version, update date, article, and maybe cache age
			#I've removed wiki status
			$tablehtml .= "<th class='rrwikidata rrseparator'>" . str_replace(' ', '&thinsp;', $admtext['mmchwversion']) . "</th>
				<th class='rrwikidata'>" . str_replace(' ', '&thinsp;', $admtext['mmchwupdate']) . "$sort</th>
				<th class='rrwikidata'>{$admtext['mmchwarticle']}</th>\n";
			#Cache age is optional
			if ($options['showcache'])
				$tablehtml .= "<th class='rrwikidata'>{$admtext['mmchwcache']}</th>\n";
		} else { #Simple report - Version, update date, article
			$tablehtml .= "<th class='rrwikidata'>{$admtext['mmchwversion']}</th>
				<th class='rrwikidata'>{$admtext['mmchwupdate']}</th>
				<th class='rrwikidata'>{$admtext['mmchwarticle']}</th>\n";
		}
		###### Wiki Summary ######
		if ($options['summary']) {
			$tablehtml .= "<th><div style='float:right;'>
				<span class='snlink summarytoggle'>{$admtext['mmcmore']}</span>
				<span class='snlink summarytoggle' style='display:none;'>{$admtext['mmcless']}</span></div>
				<div>{$admtext['mmcsummary']}</div></th>";
		}
	}


	###### Author goes here if the 'aright' flag is set
	$sort = $options['sb']=='author' ? $sortflag1 : '';
	if ($options['aright'])
		$tablehtml .= "<th class='authorcol' style='min-width:55px;'>{$admtext['mmcauthor']}$sort</th>\n";

	$tablehtml .= "</tr></thead>\n";
	echo "$tablehtml<tbody>\n";

	##### Step through the sorted mods, outputting the table rows and counting status value occurrences.
	#The mods are in the associative array $modResults[{sortkey}][{modname}] = {mod data}, where
	#{mod data} is also an associative array, as generated by (and documented in) function createResultRow.
	#That array has the following subscripts and values:
	# 4 candidate search keys - filename, modname, author, and Wiki article update date
	# 4 status values that are also candidate search keys:
	#       site1vwiki, site2vwiki, site2vsite1, and wiki.
	# and the HTML code for the mod's result table row, in the subscript 'tablerow'
	#At this point, we need to step through the $tablebody subscripts to get to a mod, and then
	#increment counts of occurrences of the author and the status values, and, finally,
	#output the HTML.

	$tablerowcount = 0;
	foreach ($modResults as $key=>$value) {
		#When sorting by anything except modname, filename, and date, add a bold line between each
		#value of the sort key.
		if ($options['sb'] != 'modname' && $options['sb'] !== 'filename' && $options['sb'] !=='date')
			echo "<tr class='rowseparator'><td colspan=20 class='rowcellseparator'></td></tr>\n";
		foreach ($modResults[$key] as $key2 => $modArray) {
			#Accumulate status counts
			#We didn't initialize these because we don't know all possible status values
			foreach (explode(',', 'site1vwiki,site2vwiki,site2vsite1,wiki') as $ss) {
				if (!empty($modArray[$ss])) {
					if (isset($rowcounts[$ss][$modArray[$ss]]))
						$rowcounts[$ss][$modArray[$ss]]++;
					else $rowcounts[$ss][$modArray[$ss]] = 1;
				}
			}
if (0) {
			if (!empty($modArray['site1vwiki'])) {
				if (isset($rowcounts['site1vwiki'][$modArray['site1vwiki']]))
					$rowcounts['site1vwiki'][$modArray['site1vwiki']]++;
				else $rowcounts['site1vwiki'][$modArray['site1vwiki']] = 1;
			}
			if (!empty($modArray['site2vwiki'])) {
				if (isset($rowcounts['site2vwiki'][$modArray['site2vwiki']]))
					$rowcounts['site2vwiki'][$modArray['site2vwiki']]++;
				else $rowcounts['site2vwiki'][$modArray['site2vwiki']] = 1;
			}
			if (!empty($modArray['site2vsite1'])) {
				if (isset($rowcounts['site2vsite1'][$modArray['site2vsite1']]))
					$rowcounts['site2vsite1'][$modArray['site2vsite1']]++;
				else $rowcounts['site2vsite1'][$modArray['site2vsite1']] = 1;
			}
			if (!empty($modArray['wiki'])) {
				if (isset($rowcounts['wiki'][$modArray['wiki']]))
					$rowcounts['wiki'][$modArray['wiki']]++;
				else $rowcounts['wiki'][$modArray['wiki']] = 1;
			}
}
			#The table row HTML has placeholders for the row count, which isn't known until the
			#complete table has been sorted.
			echo str_replace('XXXXX', ++$tablerowcount, $modArray['tablerow']),"\n";
		}
	}
	echo "</tbody></table>\n";
} #end function printResultsTable

####################################################################
#### Below the table, display counts of what we saw and what we processed.
#Display the status counts in 4 columns, like the status checkboxes in the form.
#These are the status counts of all mods we looked at, not just the mods that wound up
#in the results table. If a mod was eliminated by data in the .cfg file, that mod
#is not included here, but ALL mods that were processed by the main processing loop
#ARE here.  Status values that were sort filters are highlighted
#First, display a heading, and then define a wrapping div with overflow:auto to
#contain the left float that is applied to each column.
function printSummaryCounts($site2count) {
	global $options, $admtext, $rowcounts;

	echo '<br/><b>',$admtext['statusseen'];
	if ($options['wiki']) echo ' ' . $admtext['andwiki'];
	echo ":</b>";
	if ($options['statusconditions'])
		echo "<span class='nw rrselectedstatus'> ({$admtext['hilitecriteria']})</span>\n";
	echo "<div>\n"; #a container div for the (up to) 4 status columns
	if ($options['wiki']) {
		echo displayStatusCounts('site1vwiki');
		if ($site2count) echo displayStatusCounts('site2vwiki');
	}
	if ($site2count) echo displayStatusCounts('site2vsite1');
	echo "<div style='clear:both;'></div>";
	echo "<div>\n"; #end the container division
} #end function printSummaryCounts

###############################################
#Inputs: site1data and site2data are associative arrays that contain data from a mod's
#.cfg file on the local site, and the same mod's .cfg file on the remote site.
#The two arrays are combined with the mod's Wiki data (in the local array $wikidata)
#to produce an HTML string that can be a row in the Comparison Report's result table.
#The HTML string is returned as part of an associative array that contains elements that
# - identify the mod uniquely,
# - provide possible sort keys, and
# - contain the HTML string.
#$site1data or $site2data can be an empty string, if the current mod is not present on the site.
function createResultRow($site1data, $site2data) {
	global $rowcounts; #Counts occurrences of each status
	global $modspath, $options, $admtext, $wikidata;

	$rowcounts['all']++;
	$fileurl='';
	$authorsource = $cfgauthor = $cfgauthortext = '';

	#First, pull values from the site1 results AND site2 results.
	if (is_array($site1data)) {
		$rowcounts['site1analyzed']++;
		$primarySite = 'site1';
		$modname = $site1data['modname'];
		$modnamecompare = $site1data['modnamecompare'];
		$filename = $site1data['filename']; #The .cfg extension has not been removed.
		$fileurl = "{$options['site1url']}/$modspath/$filename";
		$primaryVersion = $site1Version = $site1data['version'];
		$primarySortableVersion = $site1SortableVersion = $site1data['sortableVersion'];
		$site1InstalledStatus = $site1data['installed'];
		if ($site1data['author']) {
			$cfgauthor = $site1data['author'];
			$cfgauthortext = $site1data['authortext'];
			if ($options['site2url']) $authorsource = '1';
			else $authorsource = 'F';
		}
		$site1Private = $site1data['private'];
		$wikipage = $site1Wikipage = $site1data['wikipage'];
		#We don't need much from the site2 results; for instance, the filename and modname
		#and their variants, were defined by site1.
		if (is_array($site2data)) {
			$rowcounts['site2analyzed']++;
			$site2Version = $site2data['version'];
			$site2SortableVersion = $site2data['sortableVersion'];
			$site2InstalledStatus = $site2data['installed'];
			if (!$cfgauthor) {
				if ($site2data['author']) {
					$cfgauthor = $site2data['author'];
					$cfgauthortext = $site2data['author'];
					$authorsource = '2';
				}
			}
			$site2Private = $site2data['private'];
			$site2Wikipage = $site2data['wikipage'];
			if (!$wikipage) $wikipage = $site2Wikipage;
			#Calculate the site2 vs. site1 status
			if ($site1Version == '??' || $site2Version == '') {
				$site2VsSite1 = '??';
			} else {
				if ($site2SortableVersion > $site1SortableVersion) $site2VsSite1 = '>site1';
				elseif ($site2SortableVersion < $site1SortableVersion) $site2VsSite1 = '<site1';
				else $site2VsSite1 = '=site1';
			}
		} else { #There was a site1 mod, but not a site2 mod.
			$site2Wikipage = $site2Private = $site2Version = $site2SortableVersion = '';
			$site2InstalledStatus = '';
		}
	}
	else { #There is no site1 data, so we have to get certain data from the site2 mod.
		$rowcounts['site2analyzed']++;
		$primarySite = 'site2';
		$modname = $site2data['modname'];
		$modnamecompare = $site2data['modnamecompare'];
		$filename = $site2data['filename'];
		$fileurl = "{$options['site2url']}/$modspath/$filename";
		$primaryVersion= $site2Version = $site2data['version'];
		$primarySortableVersion = $site2SortableVersion = $site2data['sortableVersion'];
		$site2InstalledStatus = $site2data['installed'];
		if ($cfgauthor == $site2data['author']) {
			$cfgauthortext = $site2data['authortext'];
			$authorsource = '2';
		}
		$site2Private = $site2data['private'];
		$wikipage = $site2Wikipage = $site2data['wikipage'];
		#Make sure some site1 values are cleared.
		$site1InstalledStatus = $site1Wikipage = $site1Private = $site1Version = $site1SortableVersion = '';
		$site2VsSite1 = 'mmna';
	}

	#Remove "mod" from the end of the filename comparison string
	$modnamecompare = preg_replace('#mod$#i', '', $modnamecompare);

	####Get wiki data if we are supposed to look at the wiki & the mod is not private
	#and either site has a Wikipage reference.
	if ($options['wiki'] && $wikipage && !($site1Private || $site2Private) ) {
		#GetWikiData returns a scalar value if the Wiki article cannot be found.
		$wikidata = getWikiData($wikipage);
		if (!is_array($wikidata))
			return array('tablerow'=>''); ######################EARLY RETURN ##############

		global $netmods;
		
		#Display the mod names in the function that reads the Wikipage,
		#not here
		if (false && $options['fromform'])
			echo "<span class='nw'>", ++$netmods,".$modname,</span> ";

		#If there is no cfg author, get the mod author from the Wiki data.
		if (!$cfgauthor && $options['authorisrelevant']) {
			#The author is a search criterion
			$authormatches = (stripos($wikidata['authortext'], $options['author'])!==false);
			if (($options['ayesno']=='y' && !$authormatches) ||
				($options['ayesno']=='n' && $authormatches))
			return array('tablerow'=>'');  ########## EARLY RETURN; AUTHOR DOESN'T MATCH #########
		}

		##If there's a site1 mod, calculate its status (i.e. status vs. wiki)
		if (is_array($site1data))
			$site1VsWiki = compareSiteToWiki($site1SortableVersion);
		else $site1VsWiki = '';
		#If there's a site2 mod, calculate its status v. Wiki
		if (is_array($site2data))
			$site2VsWiki = compareSiteToWiki($site2SortableVersion);
		else $site2VsWiki = '';
	} else { #Wiki was skipped - calculate site1 and site2 status based on Private and Wikipage tags
		if (!is_array($site1data)) $site1VsWiki = '';
		elseif ($site1Private) $site1VsWiki = 'mmprivate';
		elseif (!$site1Wikipage) $site1VsWiki = 'mmnolink';
		if (!is_array($site2data)) $site2VsWiki = '';
		elseif ($site2Private) $site2VsWiki = 'mmprivate';
		elseif (!$site2Wikipage) $site2VsWiki = 'mmnolink';
		$wikidata = array('version' => '', 'sortableversion' => '', 'construction'=>'');
	}

	##### The Author cell #####
	#Highlight the author field if it contains the author being searched for and if
	#we are not using the author to restrict the mods being displayed.
	#We now need to separate the author's name for the purpose of matching from the name
	#we will display, since we want to be able to display abbreviated and hyperlinked names
	$hiliteclass = '';
	#Use the author from the Wiki article only if there was no author in a cfg file.
	if (!empty($cfgauthor)) {
		$author = $cfgauthor;
		$authortext = $cfgauthortext;
	} else { #Use the Wiki author
		$author = $wikidata['author'];
		$authortext = $wikidata['authortext'];
		$authorsource = 'W';
	}
	$authorhtml = '';
	if ($options['aformat']!='none') {
		if ($author) {
			#Highlight author names that match the searchstring.
			if ($options['author'] && !$options['ayesno']) {
				if (stripos($author, $options['author'])===0) {
					#The mod's author matches the author being searched for
					$rowcounts['author']++;
					$hiliteclass = " class='installed'";
				} else $hiliteclass = " class='ready'";
			} else $hiliteclass = '';
			$authorhtml .= "<td$hiliteclass>";
			#Format the author name as first, first plus last initial, or first and last.
			#Start by breaking the name into space-sparts
			$authorparts = explode(' ', $author);
			#Always keep a middle initial in the author's name by adding it to the first name.
			if (count($authorparts)==3 && strlen($authorparts[1])==1) {
				$authorparts[0] .= ' ' . $authorparts[1];
				$authorparts[1] = array_pop($authorparts);
			}
			#Abbreviate the author if either of two values of the authorformat option is set
			if ($options['aformat']=='first')
				$authordisplay = $authorparts[0];
			elseif ($options['aformat']=='initial')
				$authordisplay = $authorparts[0] . ' ' . substr($authorparts[1],0,1);
			elseif ($options['aformat']=='none')
				$authordisplay = '';
			else $authordisplay = $author;

			$sstext = $superscript = '';
			if ($authorsource && ($options['site2url'] || $options['wiki'])) {
				$sstext = '<br>' . str_replace('XXX', $admtext["mmcsource$authorsource"], $admtext['mmcsource']);
				if ($options['showsuper'] && $authorsource<>'F')
					$superscript = "<sup>$authorsource</sup>";
			} else $superscript = $sstext = '';

			#Author popups need to expand to the left when the author column is placed on the right.
			$authorclass = $options['aright'] ? "rrauthorpop" : "rrcellpop2";
			#Now that we've formatted the name, see if we can just grab the hyperlink from
			#the full authortext and apply that hyperlink to the display name.
			if (strip_tags($authortext) == $author) {
				if ($authordisplay != $author)
					$authordisplay = str_replace($author, $authordisplay, $authortext);
				else $authordisplay = $authortext;

				if ($authordisplay == $authortext) {
					if ($sstext)
						$authorhtml .= "<div class='rrcellpop1'>$superscript$authordisplay
							<div class='$authorclass'>$sstext</div></div></td>";
					else #Just display the author, without a popup
						$authorhtml .= "$superscript$authordisplay</td>";
				} else #Display the authortext in a popup
					$authorhtml .= "<div class='rrcellpop1'>$superscript$authordisplay
						<div class='$authorclass'>$authortext$sstext</div></div></td>";
			} else { #There is more to the authortext than just the author name.
				#Add a plus sign and display the authortext in a popup
				$authorhtml .= "<div class='rrcellpop1'>$superscript$authordisplay<sup><b>+</b></sup>
					<div class='$authorclass'>$authortext$sstext</div></div></td>";
			}
		} #end if then there is an author name
		else $authorhtml .= "<td>&nbsp;</td>"; #no author
	}
	##### End build Author cell #####

	#################### Define a table data row ##########################
	#It may or may not be put into the table array (by the calling routine),
	#depending on the row restrictions (only a certain author, status, etc.).
	#The value 'XXXXX' is written to the row count column, and will be replaced after the table
	#array is sorted, and when the table array is being written to the HTML stream.
	$html = "<tr class='modline'>
		<td>XXXXX</td>\n";

	###### Modname/filename #####
	$modnameclass='';
	if (!$options['reportclass']) {
		#Print modname OR filename (as is)
		if ($options['sb'] == 'filename')
			$html .= "<td>" . empty2nbsp($filename) . "</td>\n";
		else $html .= "<td class='$modnameclass'>" . empty2nbsp($modname) . "</td>\n";
	} else { #normal report - both modname and filename, with all kinds of filename processing
		$html .= "<td class='$modnameclass'>" . empty2nbsp($modname) . "</td>\n";
		#Reformat the filename and check for modname or version# mismatches as allowed by options.
		#The function return value is a 3-element array:
		#filename (possibly abbreviated) and always without .cfg, modname error flag, version# error flag
		$filenamearray = findFilenameProblems($filename, $modnamecompare, $primaryVersion);
		$classname = $popupmessage = '';
		if ($filenamearray[0] != $filename)
			$popupmessage .= $filename . "<br>";
		if ($options['mismatch'] && ($filenamearray[1] || $filenamearray[2])) {
			$classname = 'badcfg';
			if ($filenamearray[1])
				$popupmessage .= $admtext['mmcmodnameerror'] . '<br>';
			if ($filenamearray[2])
				$popupmessage .= $admtext['mmcversionerror'];
		}
		$filename = $filenamearray[0];

		#Don't let links from outside expose the .cfg file contents
		if ($options['fromform'])
			$filenamelink = "<a href=\"$fileurl\" target='_blank'";
		else $filenamelink = $filename;

		#Are we allowed to wrap the filename?
		if ($options['wrap'])
			#Add zero-length spaces after underscores to facilitate wrapping.
			$filename = str_replace('_', '_&#8203;', $filename);
		else $classname .= ($classname ? ' ' : '') . 'nw';
		$fullfilenamelink = $filenamelink . ">$filename</a>$popupmessage";
		$filenamelink .= ">$filename</a>";

		if ($classname) $hiliteclass=" class='$classname'";
		else $hiliteclass='';
		if ($popupmessage)
			$html .= "<td$hiliteclass><div class='rrcellpop1'>$filenamelink<div class='rrcellpop2'>$popupmessage</div></div></td>";
		else
			$html .= "<td$hiliteclass>$filenamelink</td>";
	} #end if mymods report or not,
	###### end of modname/filename #####

	###### Author goes here unless the 'aright' flag is set
	if ($options['aformat']!='none' && !$options['aright'])
		$html .= $authorhtml;

	################# Border between ID columns and site1 #################
	if ($options['site2url']) $colstyle = ' rrseparator';
	else $colstyle = '';


	################## Put site1 data in the table row ##################
	####No installation status or version when coming from outside
	#unless the 'installed' flag is set."
	if ($options['fromform'] || $options['installed']) {
		if ($site1InstalledStatus) {
			if ($options['statushilite']) {
				$classname = $site1InstalledStatus;
			} else $classname = '';
			if ($site1InstalledStatus == 'ready')
				$site1InstalledStatus = 'ok2inst';
			elseif ($site1InstalledStatus == 'badcfg')
				$site1InstalledStatus='cantinst';
			$html .= "<td class='rrsite1data $classname$colstyle'>" .
				empty2nbsp($site1InstalledStatus) . "</td>";
		} else $html .= "<td class='rrsite1data$colstyle'>&nbsp;</td>";
		$colstyle = '';
	} 

	#### Site 1 Version - not in external report unless 'v' is set
	if ($options['fromform'] || $options['v']) {
		$html .= "<td class='colstyle'>" . empty2nbsp($site1Version) . '</td>';
		$colstyle = '';
	}

	### Site1 vs Wiki - The simple and full comparison reports
	if ($options['wiki'] && ($options['fromform'] || $options['v'])) {
		if ($site1VsWiki) {
			$classname = getDisplayClassForSiteVsWiki($site1VsWiki);
			if (isset($admtext[$site1VsWiki])) $displayvalue = $admtext[$site1VsWiki];
			else $displayvalue = str_replace('<', '&lt;', $site1VsWiki);
			$html .= "<td class='center rrwikidata $classname$colstyle'>" . empty2nbsp($displayvalue) . "</td>\n";
		} else
			$html .=  "<td class='rrwikidata$colstyle'>&nbsp;</td>";
		$colstyle = '';
	}

	if ($options['site2url']) {
		################## Site2 data ##################
		$colstyle = ' rrseparator';
		#First, the installation status
		if ($site2InstalledStatus) {
			if ($options['statushilite']) {
				$classname = $site2InstalledStatus;
			} else $classname = '';
			if ($site2InstalledStatus == 'ready') $site2InstalledStatus = 'ok2inst';
			elseif ($site2InstalledStatus == 'badcfg') $site2InstalledStatus='cantinst';
			$html .= "<td class='rrsite2data $classname$colstyle'>" . empty2nbsp($site2InstalledStatus) . "</td>";
		} else $html .= "<td class='rrsite2data$colstyle'>&nbsp;</td>";
		$colstyle='';

		#Site 2 version #
		$html .= "<td class='rrsite2data'>" . empty2nbsp($site2Version) . "</td>";

		#Site 2 v Site1
		if ($site2VsSite1) {
			if (isset($admtext[$site2VsSite1])) $site2VsSite1 = $admtext[$site2VsSite1];
			if ($options['statushilite']) {
				if ($site2VsSite1 == '>site1' || $site2VsSite1 == '<site1')
					$classname = 'badcfg';
				else $classname = 'ready';
			} else $classname = '';
			$html .= "<td class='rrsite2data center $classname'>" . str_replace('<', '&lt;', empty2nbsp($site2VsSite1)) . "</td>";
		} else $html .= "<td class='rrsite2data'>&nbsp;</td>";

		#Site2 v Wiki - only if the wiki has been selected
		if ($options['wiki']) {
			if ($site2VsWiki) {
				if ($options['statushilite']) {
					$classname = getDisplayClassForSiteVsWiki($site2VsWiki);
					if (isset($admtext[$site2VsWiki])) $displayvalue = $admtext[$site2VsWiki];
					else $displayvalue = str_replace('<', '&lt;', $site2VsWiki );
				}
				$html .= "<td class='center rrsite2data rrwikidata $classname'>" . empty2nbsp($displayvalue ) . "</td>\n";
			} else $html .= "<td class='rrsite2data'>&nbsp;</td>";
		}

	} #end if there is a site 2 to display

	################## Wiki columns ######################
	if ($options['wiki']) {
		#Make sure the wiki attributes have values
		foreach (explode(',', 'version,moddate,wikititle,status,wikiurl,versiontext') as $ss)
			if (empty($wikidata[$ss])) $wikidata[$ss]='';
		### The full report gets a separator
		if ($options['reportclass']) $colstyle = ' rrseparator';

		#### Wiki version - all reports
		if ($wikidata['versiontext'])
			$html .= "<td class='rrwikidata$colstyle'><div class='rrcellpop1'>{$wikidata['version']}<sup><b>+</b></sup>
				<div class='rrcellpop2'>{$wikidata['versiontext']}</div></div></td>\n";
		else #No popup
			$html .= "<td class='rrwikidata$colstyle'>{$wikidata['version']}</td>\n";
		$colstyle = '';

		####Date of Latest mod revision - all reports
		$html .= "<td class='rrwikidata right'>" . empty2nbsp($wikidata['moddate']) . '</td>';

		### Wiki article - all reports
		#Highlight Wiki article if its title is different from the mod name
		$wikititlecompare = strtolower(preg_replace('@[-_ ]@', '', $wikidata['wikititle']));
		$wikititlecompare = preg_replace('@mod$@i', '', $wikititlecompare);
		$classname = '';
		if ($wikidata['status'] == 'mmmissing') {
			$classname = ' rrstrikeoutclass';
			if ($options['statushilite']) $classname .= ' badcfg';
		} elseif ($wikidata['wikititle'] && $options['mismatch'] && 
			$wikititlecompare != $modnamecompare)
			$classname .= ' badcfg';
		$wikititle = str_replace('_', '_&#8203;', $wikidata['wikititle']);
		$html .= "<td class='rrwikidata$classname'><a href=\"{$wikidata['wikiurl']}\"
			target='_blank'>$wikititle</a>";
		#Add a superscript UC to the wiki article name if it is under construction
		if ($options['showsuper'] && !empty($wikidata['construction']))
			$html .= '<sup>UC</sup>';
		$html .= "</td>\n";

		# We don't seem to be able to read the wiki page update time any more...
		#$html .= printDateInterval($wikidata['wikitime']);

		### Cache age only in full report && cache age flag
		if ($options['reportclass'] && $options['showcache'])
			#Cache age
			$html .= "<td class='rrwikidata'>" . empty2nbsp(getDateInterval($wikidata['cachetime'])) . '</td>';

		### Mod Summary - short and long versions. ###
		if ($options['summary']) {
			$summary = str_replace('"', "'", strip_tags($wikidata['summary']));
			if (strlen($wikidata['summary'])>60)
				$summary = substr($summary, 0, 58) . 
				"<span class='summarytoggle2'>&hellip;</span><span class='summarytoggle2' style='display:none;'>" .
				substr($summary, 58) . "</span>";
			$html .= "<td class='wikisummary' title=\"$modname: $summary\">$summary</td>\n";
		}
	} ###### End Wiki Data #####

	###### Author goes here if the 'aright' flag is set
	if ($options['aformat']!='none' && $options['aright'])
		$html .= $authorhtml;

	$html .= "</tr>\n";

	#Return an array that consists of possible sort values (filename, modname, author, mod
	#revision date, and 4 representations of mod status), and restriction values (author, mod status),
	#plus the table row we just generated.
	#The parent routine use the restriction values to determine whether to add this table row
	#to the array that will become the HTML output, and will determine which sort value will be
	#used to sort the table.
	#Note: We never look at 'wikiversion' again. It's for debugging.
	#Note: We aren't displaying the Wiki status any more, but we need it when
	#we are color-coding the version comparison.
	if (empty($site2VsSite1)) $site2VsSite1='';
	if (empty($wikidata['modtime']))$wikidata['modtime']='';
	if (empty($wikidata['status']))$wikidata['status']='';
	if (empty($wikidata['version']))$wikidata['version']='';
	return array('filename'=>$filename, 'modname'=>$modname, 'author'=>$author,
		'date'=>$wikidata['modtime'],
		'site1vwiki'=>$site1VsWiki, 'site2vwiki'=>$site2VsWiki,
		'site2vsite1'=>$site2VsSite1, 'wikistatus'=>$wikidata['status'],
		'wikiversion'=>"$site1Version|$site2Version|{$wikidata['version']}",
		'tablerow'=>$html);
} #end function createResultRow

#######################################################
/* Display a column of status counts.
This function is called up to 4 times; once for each of the status arrays.
The input argument $statustype serves as
1. The base $admtext subscript for the label to be displayed at the top of a column of status values.
   (The string 'status' is added to $statustype.) and as the bas
2. A subscript in $options['statusvalues'][$statustype][{status value}] where the status
   search values are stored, and
3. A subscript in $rowcounts[$statustype][{status value}], where instances of a status value
   are stored.

Note that the cinstances of a status value are counted when a mod's results table row is being
generated, even though some mod are eliminated after that point.
As a result, the status value counts do not reflect the occurrences in the result table.
Instead, they represent the occurrences where  when $statusCounts is based on looking at mods BEFORE they were eliminated from
consideration as a result of not having the required status values in $selectedStatusValues.
As a result, $statusCounts can have status values that are not in $selectedStatusValues,
and $selectedOptions can have status values that were never seen.
*/
function displayStatusCounts($statustype) {
	global $admtext, $options, $rowcounts;

	#Use the style class .rrstatuscolumn to causes the column to float left.
	$html = "<div class='rrstatuscolumn'><b>{$admtext['mm'.$statustype.'status']}</b>";
	#First, display the counts for the selected status values, and highlight them. The value in
	#$selectedStatusValues is always 1, but we display the corresponding value from $statusCounts.
	if (isset($options['statusvalues']) && is_array($options['statusvalues'][$statustype])) {
		foreach ($options['statusvalues'][$statustype] as $key=>$value) {
			#If the status value has a translation, display the translation.
			#Otherwise, replace '<' with its entity.
			if (isset($admtext[$key])) $keyhtml = $admtext[$key];
			else $keyhtml = str_replace('<', '&lt;', empty2nbsp($key));
			$html .= "<br/><span class='rrselectedstatus'>$keyhtml: " . (+$rowcounts[$statustype][$key]) . "</span>\n";
			#Remove this entry from the status counts so we don't duplicated it
			unset ($rowcounts[$statustype][$key]);
		}
	}
	#Display the remaining status counts
	if (is_array($rowcounts[$statustype])) {
		foreach ($rowcounts[$statustype] AS $key=>$value) {
			#The values are counts. The keys are the status values.
			#If the status value has a translation, display the translation.
			#Otherwise, replace '<' with its entity.
			if (isset($admtext[$key])) $keyhtml = $admtext[$key];
			else $keyhtml = str_replace('<', '&lt;', empty2nbsp($key));
			$html .= "<br/>$keyhtml: $value\n";
		}
	}
	return "$html</div>\n";
} #End function displayStatusCounts

##############################################################
# Function to convert a set of status value checkboxes defined in the form
#     $_GET({basefieldname}field][i])=status
# into a lookup array in the form
#     $newarray[status] = 1
# Where $newarry is the return value of this function.
# $options['statusconditions'] lets lets us know whether there are any status options
# in any of the four status arrays.
function createStatusLookupArray($basefieldname) {
	global $options, $exptime;
	if (!empty($_GET[$basefieldname.'field']) && is_array($_GET[$basefieldname.'field'])) {
		#Save the status values to a string.
		$cookievalue = implode(',', $_GET[$basefieldname.'field']);

		#Save the status values to $options['statusvalues']
		foreach ($_GET[$basefieldname.'field'] AS $value) {
			$options['statusvalues'][$basefieldname][$value] = 1;
			$options['statusconditions'] = true;
		}
	} else 	$cookievalue = '(empty)';
	#Save a cookie;
	setcookie("tng_modcompare[$basefieldname]", $cookievalue, $exptime);
} #end function createStatusLookupArray

#######################################################
#Return true if a status search attribute has a value that allows us to display the current mod.
#The parameter is the array returned by createResultRow, in which the 4 status attribute names are keys;
#Exit as soon as any match is found.
function checkSelectedOptions($result) {
	#Check each of the 4 arrays of option values
	global $options;
	if (!$options['statusconditions']) return true;
	if ($options['statusvalues']['site1vwiki'][$result['site1vwiki']]) return true;
	if ($options['statusvalues']['site2vwiki'][$result['site2vwiki']]) return true;
	if ($options['statusvalues']['site2vsite1'][$result['site2vsite1']]) return true;

	return false;
} #End function checkSelectedOptions

#######################################################
#Display All Selected values for one Status Type.
#Called from the code that displays all search parameters
#Input argument: $statustype = a label; 'site1vwiki', 'site2vwiki', 'site2vsite1', or 'wiki'
#$options['statusvalues'][$statustype] is a lookup array, where the keys are the required
#status values for this status type.
function displaySelectedOptions($statustype) {
	global $options, $admtext;
	#If there are no options, just return
	if (!is_array($options['statusvalues'][$statustype]) || !count($options['statusvalues'][$statustype])) return '';
	$html = '';
	#The value is always 1, but we're looking for the Key and can't get it unless we get the value, too.
	foreach ($options['statusvalues'][$statustype] AS $key=>$value) {
		if ($html) $html .=', ';
		if (isset($admtext["mm$key"])) $keyhtml = $admtext["mm$key"];
		else $keyhtml = str_replace('<', '&lt;', $key);
 		$html .= "'" . empty2nbsp($keyhtml) . "'";
	}
	return "<br/>{$admtext['mm'.$statustype.'status']} = " . $html;
} #end function displaySelectedOptions

########################################################
# Take a Unix datetime, and, in the results table, output it relative to now in terms of
# minutes or hours or days or weeks.
# Use minutes up to 120, hours up to 72, days up to 28, and weeks beyond that.
function getDateInterval($dt) {
	if ($dt=='') $interval='';
	elseif ( $dt < 0) $interval = '??';
	else {
		$interval = time() - $dt;
		if ($interval < 7200 ) $interval = -(floor($interval/60)) . 'm';
		elseif ($interval < (86400*3)) $interval = -(floor($interval/3600)) . 'h';
		elseif ($interval < (604800*4)) $interval = -(floor($interval/86400)) . 'd';
		else $interval = -(floor($interval/604800)) . 'w';
	}
	return $interval;
} #end function getDateInterval

########################################################
# getWikiData reads information about a mod from its Wiki article
# Input argument:
# $wikititle - The title of a wiki article.
# $cfgauthor - The authorname from the cfg file (tells us we don't need to check the Wiki authorname)
# Return value - an associative array whose local name is $wikidata, with these keys:
# 'wikititle' & 'wikiurl' - References to the Wiki article
# 'status' - The status of the wiki page
# 'modstatus' - Status value for the mod, for cases where the wikiurl doesn't find a wiki page.
# 'linkclass' - A highlighting class for the Wiki hyperlink.
# 'versiontext' - Full text of the Wiki sidebar's Version parameter.
# 'version' - The version number that will represent the Wiki article - can be a message such as '??'
# 'sortableVersion' - A version number padded with spaces so it can be compared to others.
# 'authortext' - Full text from the "Author(s)" parameter in the Wiki sidebar.
# 'author' - The author's name, extracted from the beginning of 'authortext'
function getWikiData($wikititle) {
	global $wikicache, $admtext, $options;
	global $rowcounts; #counters displayed below the results table

	$fromCache = $fromServer = false; #Where we ultimately get the wiki data
	if ($wikititle) {
		if (empty($wikicache[$wikititle]['modtime'])) $wikicache[$wikititle]['modtime']='';
		#When this mod's cache was updated
		$cachetime = empty($wikicache[$wikititle]['cachetime']) ? '' : $wikicache[$wikititle]['cachetime'] ;
		#When the wiki article was updated, according to the Wiki server
		# can't read it any more...$wikitime = $wikicache[$wikititle]['wikitime'];
		#The Mod Update date, from the text of the Wiki page
		$modtime = $wikicache[$wikititle]['modtime'];
		$wikiurl = "https://tng.lythgoes.net/wiki/index.php?title=" . str_replace(' ','_',$wikititle);
		if ($options['cachemethod'] == 1 ) {
			#Read the article even if there is cached data
			$wikidata = readWikiPage($wikiurl, $wikititle, $cachetime);
			$fromServer = true;
		}
		elseif (!$cachetime) {
			#No cache for this mod
			$wikidata = readWikiPage($wikiurl, $wikititle, 0);
			$fromServer = true;
		}
		elseif ($options['cachemethod'] == '3') {
			#Fastest possible method => use cached data even if it has expired.
			$wikidata = $wikicache[$wikititle]; #Get data from cache
			$fromCache = true;
		}
		else { #Check the update date of the wiki article to see if it has changed since the article was cached.
			#but the curl library isn't working.
			if ($modtime) {
				#There is a cached Wiki datetime, so ask the Wiki for an update.
				$newwikitime = getUrlUpdateTime($wikiurl);
				if ($wikitime < $newwikitime) {
					#The Wiki article has been updated
					$wikidata = readWikiPage($wikiurl, 0, $wikititle);
					$fromServer = true;
				} elseif ($newwikitime > 0) {
					#The Wiki page has not been updated, so use the cache.
					$wikidata = $wikicache[$wikititle];
					$fromCache = true;
				}
			} //end of then clause that says we can check the Wiki for an update
			if (empty($wikidata)) {
				#We don't have our wikidata yet...
				if ($cachetime < $options['cachecutoff']) {
					#The cache has expired; Read the Wiki article.
					$wikidata = readWikiPage($wikiurl, $wikititle, $cachetime);
					$fromServer = true;
				} else { #The cache is good, so use it.
					$wikidata = $wikicache[$wikititle];
					$fromCache = true;
				}
			} #end if there is an acceptable cachetime in the cache
		} # End if cache options - we now have $wikidata one way or another
	} else {
		$wikidata['status'] = $wikidata['modstatus'] = 'mmnolink';
	} #end if there is a wikiurl
	if ($fromCache) {
		$rowcounts['wikicache']++;
	} elseif ($fromServer) {
		$rowcounts['wikiserver']++;
		$wikicache[$wikititle] = $wikidata;
	}
	return $wikidata;
} #end function getWikiData

########################################################
# Use the file_get_contents to read the Wiki page at $wikiurl.
# $wikiurl - URL to read
# $cachetime - serves as a flag to indicate whether this article has been cached
# $wikititle - The article name
#Return the array $wikidata
function readWikiPage($wikiurl, $wikititle, $cachetime) {
	@ini_set( "allow_url_fopen", "1" ); #Make sure that this PHP setting is set
	global $rowcounts, $wikicache, $admtext, $options;
	$wikihtml = @file_get_contents($wikiurl);
	#See if it is a "No such page" page.
	if (!$wikihtml || strpos($wikihtml, 'There is currently no text in this page') ||
	  strpos($wikihtml, 'You have followed a link to a page that does not exist.')) {
		#There is a wiki URL, but no wiki page.
		$wikidata = array( 'status' => 'mmmissing',
			'modstatus' => 'mmmissing', 'construction'=>'');
	} else { #There is a wiki page.
		$rowcounts['wikiread']++; #Count wiki articles read
		#Parse the Wiki page to get the required data.
		$wikidata = parseWikiPage($wikihtml);
	} #end if we were able to find the wiki page
	$wikidata['wikititle'] = $wikititle;
	$wikidata['wikiurl'] = $wikiurl;
	if ($options['fromform']) echo "$wikititle, ";
	$wikidata['cachetime'] = time();
	$wikidata['cachedate'] = date('jMY H:i:s', $wikidata['cachetime']);
	return $wikidata;
} #end function readWikiPage

######################################################
#Function to gather information from a Wiki Article about a mod.
#We have confirmed that it appears to be a valid article.
#Input arguments:
# $wikihtml - The HTML code of the article
#Return value
# The array we build in the local variable $wikidata
function parseWikiPage($wikihtml) {
	global $options, $admtext;
	$wikidata = array('version' => '', 'sortableversion' => '', 'construction'=>'');

	#First, get the version number from the Wiki page
	$versiontext = getSubstring($wikihtml, "Latest&#160;Mod</th> <td>", '</td>');
	if ($versiontext == 'no match' || !trim($versiontext,"/n/r ")) {
		#No Wiki version in the Wiki article
		$wikidata['version'] = 'mmnover';
		$wikidata['status'] = 'mmnover';
	} else { #There appears to be some value in the Mod Version Number field
		#See if there are version #'s in the string
		$matchcount = preg_match_all(VERSIONPATTERN, $versiontext, $matches);
		if (!$matchcount) { #no valid version number
			$wikidata['version'] = '??';
			$wikidata['status'] = 'mmnover';
		} else {
			#Pad the beginning of each version # and the 4th part as necessary, 
			#so we can compare version #s.
			$vresults = getHighestSortableVersion($matches);
			foreach ($vresults as $sortable=>$version) {
				$wikidata['version'] = $version;
				$wikidata['sortableversion'] = $sortable;
			}
			if ($matchcount>1) {
				$wikidata['status'] = 'mm1ver';
			} else {
				#Single version # - Remove the version text if it matches the version #, ignoring the v.
				$wikidata['status'] = 'mmok';
				$v = str_replace($wikidata['version'], '', $versiontext);
				if ($v=='' | $v=='v' | $v=='V')
					$versiontext = '';
			} #end if >1 or 1 version #
		} #end if there is at least 1 version # in the version# field
		$wikidata['versiontext'] = $versiontext;
	} #end if we found something in the version# field

	## Determine whether there is an Under Construction message
	if (strpos($wikihtml, 'Ambox_construction.png'))
		$wikidata['construction'] = 'Y';
	else $wikidata['construction'] = '';

	##### Get the Author from the the Wiki page sidebar.#####
	#The mod template field called "Author(s)" contains marked-up text that
	#should begin with the primary author - perhaps in a hyperlink.
	#We'll capture the full text of that field, and try to identify the author
	$authortext = getSubstring($wikihtml, "Author\(s\)</th> <td>", '</td>');
	if ($authortext != 'no match') {
		#Replace root-relative links defined in the wiki page with full links
		$authortext = str_replace("href=\"/", "href=\"https://tng.lythgoes.net/", $authortext);
		$wikidata['authortext'] = $authortext;
		#If the full text begins with a hyperlink, then assume that the hyperlink text is the author.
		if (strtolower(substr($authortext,0,2))=='<a') {
			$authorlink = substr($authortext, 0, stripos($authortext, '</a>')+4);
			$wikidata['authorurl'] = getUrlFromHyperlink($authorlink);
			$wikidata['author'] = strip_tags($authorlink);
		} else {
			#Get the first 3 words of the authortext, stripped of all tags, and try to find a name
			$words = explode(' ', strip_tags($authortext), 4);
			#Get rid of everything following the 3 pieces
			if (count($words)==4) array_pop($words);
			#Assume that the 2nd & 3rd words are not part of the name if they don't start
			#with a capital letter.
			if (count($words)>1 && !ctype_upper(substr($words[1], 0, 1)))
				#Get rid of words 2 and 3; we're left with 1 word.
				$words = array($words[0]);
			elseif (count($words)==3) {
				if (!ctype_upper(substr($words[2], 0, 1)))
					array_pop($words); #We're left with just 2 words.
				#Remove trailing commas, semicolons, and periods from the 2nd word, and
				#see if it is just a middle initial.
				$x = $words[1];
				$words[1] = trim($words[1], ',;.');
				#Punctuation on the 2nd word means that the 3rd word isn't in the name
				if (strlen($x)>1 && $x!=$words[1])
					array_pop($words);
			}
			#Remove punctuation from the end of the last name
			$words[count($words)-1] = trim($words[count($words)-1],',;.');
			$wikidata['author'] =  implode(' ',$words);
		} #End if author has a hyperlink or not
	} #end is there author text

	## Get the update date. To be able to sort the dates in the Revision History, and to have consistent
	#display dates, use the function strtotime to convert whatever date we find  into an internal time,
	#and then format that time.
	$latestmodtime = 0;  #internal format
	#First try the sidebar update date
	#Use pregmatch all only because the code that sorts dates in the Revision history uses it.
	$sidebardate = getSubstring ($wikihtml, '>Mod Updated</th> <td>', '</td>');
	if (preg_match_all('@([\d]{1,2})[ -]?([a-z]+)[ -]([\d]{2,4})@i', $sidebardate, $matches)) {
		$latest = 0;
		$latestmodtime = strtotime($matches[0][0]);
	} else {
		#Try to get the latest update date from the Revision History section of the page.
		#Grab the HTML code starting after the ID, through the next </table> tag.
		$history = getSubstring($wikihtml, 'id="Mod_Change_History"', '</table>');
		if ($history == 'no match')
			$history = getSubstring($wikihtml, 'id="Revision_History"', '</table>');
		#We don't want a subsquent page heading inside that history
		if (!strpos($history,'class="mw-headline"')) {
			if (preg_match_all('@([\d]{1,2})[ -]?([a-z]+)[ -]([\d]{2,4})@i', $history, $matches)) {
				for ($ii=0; $ii<count($matches[0]); $ii++) {
					$timeval = strtotime($matches[0][$ii]);
					if ($timeval > $latestmodtime) {
						$latest = $ii;
						$latestmodtime = $timeval;
					}
				}
			}
		}
	}
	if ($latestmodtime) {
		$wikidata['modtime'] = $latestmodtime;
		$wikidata['moddate'] = date("dMY", $latestmodtime);
		#Save the latest update time to compare with the cache.
	}
	##### Get the Mod Summary from the Wiki page sidebar.#####
	$wikidata['summary'] = getSubstring($wikihtml, "Summary</th> <td>", '</td>');

	return $wikidata;
} #end function parseWikiPage

######################################################
#Function to format the mod fileame for display by comparing it to the modname and mod vers #.
#Return a 3-element array: filename, filename error (T/F), and version error (T/F)
#If $options['hideversion'] is true, then a matching version # will be stripped from the filename.
function findFilenameProblems($filename, $sortablemodname, $modversion) {
	global $options; #Mod Manager options
	$modnameError = $versionError = '';
	$filenameParts = getCfgFilenameParts($filename);
	$fileModname = $filenameParts[1]; #Modname part of filename
	#Part 2 is '_v'
	$sortablefilename = strtolower(str_replace('-', '', str_replace('_', '', str_replace(' ', '', $fileModname))));
	#Remove "mod" from the end of the sortable mod name and sortable filename
	$sortablefilename = preg_replace('#mod$#i', '', $sortablefilename);
	$modnameError = ($sortablemodname != $sortablefilename);

	#The file version should match the mod version (from within the file)
	$fileVersion = $filenameParts[3]; #Version # part of filename
	$versionError = ($fileVersion != strtolower($modversion));
	#If there is not a version error, we may want to omit the version # from the filename
	#We always omit .cfg
	if ($options['reportclass'] && $options['fromform'] && $options['hideversion'] && !$versionError)
		$filename = $fileModname . '&hellip;'; #Omit version #
	else $filename = $fileModname . strtolower($filenameParts[2]) . $fileVersion;
	return array($filename, $modnameError, $versionError);
} #end function findFilenameProblems

###############################################
#Separate a mod filename (with no extension) into its constituent parts
#Returns an array with 4 elements - modname, separator ('_v|V'), version
function getCfgFilenameParts($filename) {
	if (preg_match('@^(.*)(_v)(.*)\.cfg$@i', $filename, $matches))
		return $matches;
	#Try it again without the v
	if (preg_match('@^(.*)(_*)(\d+\..*)\.cfg?$@i', $filename, $matches))
		return $matches;
	#The RE match didn't work, so just return the filename without extension in element 1 of the array.
	return array('',pathinfo($filename, PATHINFO_BASENAME),'','');
} #end function getCfgFilenameParts

################################################################
# When displaying a cell value, make sure it is not empty, and replace spaces
# with the thin space entity.
function empty2nbsp($cellvalue) {
	if ($cellvalue=='') return '&nbsp;';
	else return str_replace(' ', '&thinsp;', $cellvalue);
} #end function

#########################################################
# Convert a parsed version # so that it can be sorted.
# The input parameter is a match array from preg_replace_all.
# The first 3 dot-parts are the first substring match.
# The numeric part of the 4th dot-part is the second substring match.
# There could be a suffix in the third substring match.
# Return the highest sortable version number.
function getHighestSortableVersion($matches) {
	$versions = array();
	for ($ii=0; $ii<count($matches[0]); $ii++) {
		#separate the first 3 parts of the version number
		$tngver = $matches[1][$ii];
		$tngverparts = explode('.', $tngver);
		#Add a zero to make the first version#-part 2 digits.
		if ($tngverparts[0]<10) $tngver = "0$tngver";
		#Pad the numeric part of the mod's version #
		$lastnumber = sprintf('%03d',$matches[2][$ii]);
		#Add the padded numeric version # and any suffix
		$sortable = $tngver . $lastnumber . $matches[3][$ii];
		$versions[$sortable] = $matches[0][$ii];
	}
	ksort($versions); #Sort in reverse order
	return $versions;
} #end function getHighestSortableVersion

###############################################
# readCfgFileDataFromSite - Invoke an Ajax module on $siteurl. It returns a
# JSON string with data from all mods on the site that meet the criteria
# specified in querystring parameters within $siteurl.
# The server-side module converts the JSON string into an array with
# two subscripts:
#  - numfiles - Kust the number of mod files that were found
#  - modlist = An array containing data about each mod.
#    The mod list is a numbered array (a vector) sorted by modname,
#    where each element is an associative array with the subscripts
#    'modnamecompare', 'modname', 'version', 'versionnum', 'filename',
#    'wikipage', 'installed', 'author', 'private'
function readCfgFileDataFromSite($siteurl) {
	#Global variables that determine if the mod matches a search criterion.
	global $options, $rowcounts;

	$querystring = '';
	if ($options['authorisrelevant'])
		$querystring .= "searchauthor=". urlencode($options['author']) . "&authoryesno={$options['ayesno']}&";
	if ($options['sinclude'])
		$querystring .= "searchinclude=" . urlencode($options['sinclude']) . '&';
	if ($options['sexclude'])
		$querystring .= "searchexclude=" . urlencode($options['sexclude']) . '&';
	if ($options['private'])
		$querystring .= "searchprivate=1";

	$url = $siteurl . "rrmodcompare_modlist.php?$querystring";
	$jsondata = file_get_contents($url);
if (!empty($_GET['dump'])) {
	echo nl2br(htmlentities(str_replace('},{',"},\n{",$jsondata)));
}
	$results = json_decode($jsondata, true);

	if ($results['numfiles'] == -1)
		echo "<h3 style='color:brown;margin-bottom:0;'>The mods list could not be read from {$results['$url']}</h3>";
	return $results;
} #end function readCfgFileDataFromSite

####################################################################
# Get the URL of a hyperlink
function getUrlFromHyperlink($hyperlink) {
	if (stripos($hyperlink, 'href="')!==false) $quotetype = '"';
	elseif (stripos($hyperlink, "href='") !==false) $quotetype = "'";
	else return '';
	if (preg_match('@href=' . $quotetype . '(.*)' . $quotetype .'@', $matches))
		return $matches[1];
	else return '';
} #End function getUrlFromHyperlink

###################################################################
# Extract an HTML snippet from an HTML string, using regular expression searches
# to find HTML text that comes before the snippet and HTML text that comes
# after the snippet.
function getSubstring($html, $startsearch1, $endsearch1='') {
	if (!$endsearch1) {
		if (!preg_match("@^<([a-z]*).*$@", $startsearch1, $matches))
			return "ERROR: Can't form search end string";
		$endsearch1 = '</' + $matches[1] .'>';
	}
	$pattern = "@$startsearch1(.*)$endsearch1@Us"; #s means '.' matches \n
	$x = preg_match($pattern, $html, $matches);
	if (!$x) return 'no match';
	return $matches[1];
} #end function getSubstring

###################################################################
# Write the entire cache to the cache file, overwriting the file.
# The function var_export creates PHP code that builds an array.
# After using var_export to save the cache to the file, we can read it
# as an Include file.
function saveCacheFile($wikicache) {
	global $extspath; #The wiki Cache file is stored in the extensions folder.
	$fhandle = fopen("$extspath/wikicache.php",'wt');
	fwrite($fhandle,"<?php\n// wikicache.php\n");
	fwrite($fhandle,"// This file is used by admin_modconfig.php, and contains data cached from Wiki articles.\n");
	fwrite ($fhandle, '$wikicache = ' . var_export($wikicache,true) . ";\n");
	fclose($fhandle);
} #end function saveCacheFile

##################################################################
# Ask the Wiki server when a page on the server was updated
# The input parameter is an URL, not an article name.
function getUrlUpdateTime($url) {
return 0;
   $mtime = microtime();
   $mtime = explode(" ",$mtime);
   $mtime = $mtime[1] + $mtime[0];
   $starttime = $mtime;
	$ch = curl_init($url);
	curl_setopt($ch,CURLOPT_FILETIME,true);
	curl_setopt($ch,CURLOPT_FOLLOWLOCATION,true);
	curl_setopt($ch,CURLOPT_NOBODY,true);
	curl_setopt($ch,CURLOPT_RETURNTRANSFER,true);
	curl_exec($ch);
	if (curl_errno($ch)) {
		echo "Error ",curl_errno($ch);
		return 0;
	}
	$xx = curl_getinfo($ch,CURLINFO_FILETIME);
   $mtime = microtime();
   $mtime = explode(" ",$mtime);
   $mtime = $mtime[1] + $mtime[0];
   $endtime = $mtime;
   $totaltime = ($endtime - $starttime);
   global $grandtotaltime;
   $grandtotaltime +=$totaltime;
	return $xx;
} #end function getUrlUpdateTime

##############################################################################
# Compare a site version value (in sortable format) to the Wiki status and version number.
# This is done up to twice for each Wiki file we are reading - once for site 1 and once for site 2
function compareSiteToWiki($siteVersion) {
	global $wikidata, $admtext, $options;

	if (!$siteVersion)
		$comparison = '';
	elseif ($wikidata['status'] == 'mmmissing')
		$comparison = 'mmna';
	elseif ($wikidata['status'] == 'mmnolink')
		$comparison = 'mmnolink';
	elseif ($wikidata['status'] == 'mmnover')
		$comparison = 'mmna';
	elseif ($siteVersion == '??' || $wikidata['sortableversion'] == '??')
		$comparison = 'mmna';
	elseif ($wikidata['sortableversion'] =='')
		$comparison = 'mmna';
	else {
		$x = strcmp($siteVersion, $wikidata['sortableversion']);
		if ($x < 0) $comparison = '<wiki';
		elseif ($x > 0) $comparison = '>wiki';
		#elseif ($options['fromform']) $comparison = '=wiki';
		else $comparison = '=wiki';
	}
	return $comparison;
} #end function compareSiteToWiki

#########################################################################
# After compareSiteToWiki is called to determine the status value, this function is called
# to determine the style class that will display that status value in the results table.
# Note that the style class names are defined by the core Mod Manager product, and are based on
# based on Installation Status values.
# - badcfg is pink
# - installed' is green
# - 'partinst' is orange, and
# - 'ready' is either white or the class
function getDisplayClassForSiteVsWiki($sitestatus) {
	global $options;
	if (!$options['statushilite']) return '';
	if ($sitestatus=='<wiki' || $sitestatus=='mmmissing')
		$classname = 'badcfg';
	elseif ($sitestatus == '>wiki')
		$classname = 'installed';
	elseif ($sitestatus == '??' || $sitestatus == 'mmna')
		$classname = 'partinst';
	else $classname = 'ready';
	return $classname;
} #end function getDisplayClassForSiteVsWiki

##################################################################
#This function defines the Mod Manager tab menu for all Mod Manager programs.
#Identical copies of this function are placed in each Mod Manager program.
function set_horizontal_tabs( $show_analyzer = NO, $show_developer = NO, $show_updates = NO  ) {
   global $admtext;

   $modtabs = array();
   $modtabs[0] = array(1, "admin_modhandler.php", $admtext['modlist'],"modlist");
   $modtabs[1] = array(1,"admin_showmodslog.php",$admtext['viewlog'],"viewlog");
   $modtabs[2] = array(1,"admin_modoptions.php",$admtext['options'],"options");
   if ( $show_analyzer == YES)
      $modtabs[3] = array(1,"admin_analyzemods.php",$admtext['analyzefiles'],'files');
   if ( $show_developer == YES)
      $modtabs[4] = array(1,"admin_modtables.php",$admtext['parsetable'],'parser');
   if ( $show_updates == YES)
      $modtabs[5] = array(1,"admin_modupdates.php",$admtext['recommendedfixes'],'updates');
	if (file_exists('rrinnermodmenu_newtab.php'))
		include_once('rrinnermodmenu_newtab.php');
	if (!function_exists('rrimmNewTab')) {function rrimmNewTab($modname){}}
	$modtabs['rrmmc1'] = array(1,"rrmodcompare_form.php", $admtext['mmcompare'].rrimmNewTab('Mod Manager Compare'), "compare");
	if (file_exists('rrmodcheckfiles.php'))
		$modtabs['rrmmcf'] = array(1,"rrmodcheckfiles.php",$admtext['rrmmcf-checkfiles'].rrimmNewTab('Mod Manager Check Files'), 'check');
	if (file_exists('rrmodutilities.php'))
		$modtabs['rrmmu1'] = array(1,"rrmodutilities.php", $admtext['rrmu-utilities'].rrimmNewTab('Mod Manager Utilities'), 'utilities');
   return $modtabs;
} #End function set_horizontal_tabs

################ end of PHP functions #######################
?>